use std::collections::HashMap;

use once_cell::sync::OnceCell;

use crate::util;

const VENT_WIDTH: usize = 7;

#[derive(Clone, Copy, PartialEq, Eq, Debug)]
pub enum Jet {
    Left,
    Right,
}

impl Jet {
    pub fn shift(&self) -> isize {
        match &self {
            Jet::Left => 1,
            Jet::Right => -1,
        }
    }
}

pub fn parse_jets(input: &String) -> Vec<Jet> {
    assert_eq!(input.lines().count(), 1);
    input
        .chars()
        .filter_map(|c| match c {
            '<' => Some(Jet::Left),
            '>' => Some(Jet::Right),
            '\n' => None,
            _ => panic!("Unknown jet {}", c),
        })
        .collect()
}

#[derive(Clone, Copy, Debug)]
pub enum BlockKind {
    Dash = 0,
    Plus,
    L,
    Pipe,
    Square,
}

impl BlockKind {
    pub fn from_num(n: usize) -> BlockKind {
        match n {
            0 => BlockKind::Dash,
            1 => BlockKind::Plus,
            2 => BlockKind::L,
            3 => BlockKind::Pipe,
            4 => BlockKind::Square,
            _ => panic!("Unknown block kind {}", n),
        }
    }

    pub fn height(&self) -> usize {
        match self {
            BlockKind::Dash => 1,
            BlockKind::Plus => 3,
            BlockKind::L => 3,
            BlockKind::Pipe => 4,
            BlockKind::Square => 2,
        }
    }

    pub fn width(&self) -> usize {
        match self {
            BlockKind::Dash => 4,
            BlockKind::Plus => 3,
            BlockKind::L => 3,
            BlockKind::Pipe => 1,
            BlockKind::Square => 2,
        }
    }

    pub fn vent_rows(&self) -> &'static Vec<VentRow> {
        static ROWS: OnceCell<Vec<Vec<VentRow>>> = OnceCell::new();
        let rows = ROWS.get_or_init(|| {
            vec![
                vec![0b1111],              // Dash
                vec![0b010, 0b111, 0b010], // Plus
                vec![0b001, 0b001, 0b111], // L
                vec![0b1; 4],              // Pipe
                vec![0b11, 0b11],          // Square
            ]
        });
        &rows
            .get(*self as usize)
            .unwrap_or_else(|| panic!("Unknown BlockKind: {:?}", self))
    }
}

const N_BLOCK_TYPES: usize = 5;

pub struct Block {
    kind: BlockKind,
    pos: util::Coord,
}

impl Block {
    pub fn spawn(&self, vent: &mut Vec<VentRow>) {
        for (i, row) in self.kind.vent_rows().iter().rev().enumerate() {
            vent[self.pos.0 + i] |= *row << self.shift_dist();
        }
    }

    pub fn row_range(&self) -> std::ops::Range<usize> {
        let row = self.pos.0;
        row..(row + self.kind.height())
    }

    pub fn can_push(&self, vent: &Vec<VentRow>, jet: Jet) -> bool {
        if (jet == Jet::Left && self.pos.1 == 0)
            || (jet == Jet::Right && self.pos.1 + self.kind.width() >= VENT_WIDTH)
        {
            return false;
        }

        let shift = self.shift_dist() as isize + jet.shift();
        for (i, row) in self.row_range().rev().enumerate() {
            let bitmask = self.kind.vent_rows()[i] << shift;
            if vent[row] & bitmask != 0 {
                return false;
            }
        }
        true
    }

    pub fn can_fall(&self, vent: &Vec<VentRow>) -> bool {
        if self.pos.0 == 0 {
            return false;
        }

        for (i, row) in self.row_range().rev().enumerate() {
            let bitmask = self.kind.vent_rows()[i] << self.shift_dist();
            if vent[row - 1] & bitmask != 0 {
                return false;
            }
        }

        true
    }

    pub fn shift_dist(&self) -> usize {
        VENT_WIDTH - self.pos.1 - self.kind.width()
    }
}

type VentRow = u8;

const CYCLE_SIZE: usize = 5;

pub fn do_moves(jets: &Vec<Jet>, n_blocks: usize) -> usize {
    let mut block_i = 0;
    let mut current_block: Option<Block> = None;
    let mut vent = vec![0; 4];

    // {jet_i: {round: (height, block_i)}}
    let mut block_cycles: HashMap<usize, HashMap<usize, (usize, usize)>> = HashMap::new();
    let mut skipped_height: Option<usize> = None;

    let mut round = 0;
    while block_i <= n_blocks {
        let jet_i = round % jets.len();

        if current_block.is_none() {
            if block_i % N_BLOCK_TYPES == 0 {
                if skipped_height.is_none() {
                    let entry = block_cycles.entry(jet_i).or_insert_with(HashMap::new);
                    entry.insert(round, (vent.len() - empty_rows(&vent), block_i));
                    if let Some((cycle_height, cycle_blocks)) = find_cycle(entry, CYCLE_SIZE) {
                        println!(
                            "Found cycle ({} height per {} blocks) after {} rounds!",
                            cycle_height, cycle_blocks, round
                        );

                        let remaining_blocks = n_blocks - block_i;
                        let cycles = remaining_blocks / cycle_blocks;
                        block_i += cycles * cycle_blocks;
                        skipped_height = Some(cycles * cycle_height);

                        println!(
                            "Skipping {} blocks, {} height",
                            cycles * cycle_blocks,
                            cycles * cycle_height
                        );
                    }
                }
            }
            current_block = Some(spawn_block(&mut vent, block_i % N_BLOCK_TYPES));
            // println!("Vent as block {} begins falling:", block_i);
            // print_vent(&vent, current_block.as_ref());
            // println!("");
            block_i += 1;
        }
        let block = current_block.as_mut().unwrap();

        // Push left/right
        let jet = jets[jet_i];
        if block.can_push(&vent, jet) {
            let (_, ref mut col) = block.pos;
            *col = ((*col as isize) - jet.shift()) as usize;
        }

        // Fall
        if block.can_fall(&vent) {
            let (ref mut row, _) = block.pos;
            *row -= 1;
        } else {
            block.spawn(&mut vent);
            current_block = None;
        }
        round += 1;
    }

    vent.len() - empty_rows(&vent) + skipped_height.unwrap_or(0)
}

const SPAWN_OFFSET_X: usize = 2;
const SPAWN_OFFSET_Y: usize = 3;

fn spawn_block(vent: &mut Vec<VentRow>, block: usize) -> Block {
    let block = BlockKind::from_num(block);

    let mut empty_rows = empty_rows(vent);
    let diff = (block.height() as isize + 3) - (empty_rows as isize);
    if diff > 0 {
        vent.append(&mut vec![0; diff as usize]);
        empty_rows += diff as usize;
    }

    let row = vent.len() - empty_rows + SPAWN_OFFSET_Y;
    let col = SPAWN_OFFSET_X;

    assert!(row + block.height() - 1 < vent.len());

    let block = Block {
        kind: block,
        pos: (row, col),
    };

    block
}

fn empty_rows(vent: &Vec<VentRow>) -> usize {
    vent.iter()
        .rev()
        .position(|r| *r != 0)
        .unwrap_or_else(|| vent.len())
}

pub fn print_vent(vent: &Vec<VentRow>, block: Option<&Block>) {
    let block_rows = block.map(|b| b.kind.vent_rows());
    let block_range = block.map(|b| b.row_range());
    let block_range = block_range.as_ref();
    for (r, vent_row) in vent.iter().enumerate().rev() {
        print!("|");
        let block_collision = if block.is_some() && block_range.unwrap().contains(&r) {
            block_rows.unwrap()[block_range.unwrap().end - r - 1] << block.unwrap().shift_dist()
        } else {
            0
        };
        for i in (0..VENT_WIDTH).rev() {
            let col = 1 << i;
            let c = if (block_collision & col) != 0 {
                '@'
            } else if (vent_row & col) == 0 {
                '.'
            } else {
                '#'
            };
            print!("{}", c);
        }
        print!("|\n");
    }
    println!("+-------+");
}

fn find_cycle(
    jet_heights: &HashMap<usize, (usize, usize)>,
    n_cycles: usize,
) -> Option<(usize, usize)> {
    if jet_heights.len() < n_cycles {
        return None;
    }

    for (init_round, (init_height, init_blocks)) in jet_heights {
        'next_loop: for (next_round, (next_height, next_blocks)) in jet_heights {
            if init_round == next_round || *next_height < *init_height {
                continue;
            }

            let cycle = next_round - init_round;
            let cycle_height = next_height - init_height;
            let cycle_blocks = next_blocks - init_blocks;
            for n in 2..=n_cycles {
                let round = init_round + n * cycle;
                let entry = jet_heights.get(&round);
                if let Some((height, blocks)) = entry {
                    let expected_height = init_height + n * cycle_height;
                    let expected_blocks = init_blocks + n * cycle_blocks;
                    if *height != expected_height || *blocks != expected_blocks {
                        continue 'next_loop;
                    }
                } else {
                    continue 'next_loop;
                }
            }

            return Some((cycle_height, cycle_blocks));
        }
    }

    None
}
